import abc

import flask


# =====
class ApiError(Exception):
    def __init__(self, code, message, result=None):
        super(ApiError, self).__init__()
        self.code = code
        self.message = message
        self.result = result


class Resource(metaclass=abc.ABCMeta):
    name = "<Resource>"
    dynamic = False
    methods = ("GET",)
    docstring = None

    def handler(self, **kwargs):
        try:
            (result, message) = self.process_request(**kwargs)
            return {
                "status":  "ok",
                "message": message,
                "result":  result
            }
        except ApiError as err:
            result = {
                "status":  "error",
                "message": err.message,
                "result":  err.result,
            }
            return (result, err.code)
        except Exception as err:
            message = "{}: {}".format(type(err).__name__, err)
            if hasattr(err, "__module__"):
                message = "{}.{}".format(err.__module__, message)
            result = {
                "status": "error",
                "message": message,
                "result": None,
            }
            return (result, 500)

    @abc.abstractmethod
    def process_request(self, **kwargs):
        raise NotImplementedError


def get_url_for(resource_class, **kwargs):
    return flask.request.host_url.rstrip("/") + flask.url_for(resource_class.name, **kwargs)



import sys
import os
import importlib
import traceback
import threading

from contextlog import get_logger

from ulib.validatorlib import ValidatorError
from ulib.validators.python import valid_object_name


# =====
_ATTR_EXPOSED = "_powny_exposed"


# =====
def expose(method):
    setattr(method, _ATTR_EXPOSED, True)
    return method


class Loader:
    """
        Loader() обеспечивает загрузку и перезагрузку пакета с правилами.
        При вызове get_exposed(), он удаляет все модули, __file__ которых начинается с prefix.
        Затем, в sys.path добавлется prefix/head и происходит загрузка всех пакетов и модулей
        из этого каталога, рекурсивно.
        Далее, все функции, находящиеся в этих модулях, анализируются с помощью набора фильтров
        в group_by (вида {"key": lambda func: True}), и группируются по указанным ключам.
        Операция замены модулей не является атомарной и функции, которые во время обновления
        модулей обратились к оным, потерпят сбой. FIXME: исправить это.
    """

    _lock = threading.Lock()

    def __init__(self, prefix, group_by=None):
        self._prefix = prefix
        self._group_by = group_by
        self._cache = {}
        self._thread = None

    def get_exposed(self, head):
        while True:
            if head in self._cache:
                return self._cache[head]

            if not self._lock.acquire(blocking=False):
                self._lock.acquire()
                self._lock.release()
                continue

            self._thread = threading.current_thread()
            try:
                _remove_modules_unsafe(self._prefix)
                (exposed_methods, errors) = _get_exposed_unsafe(os.path.join(self._prefix, head))
                if self._group_by is None:
                    self._cache[head] = (exposed_methods, errors)
                else:
                    methods = {sub: {} for (sub, _) in self._group_by}
                    for (name, method) in exposed_methods.items():
                        for (sub, test) in self._group_by:
                            if test(method):
                                methods[sub][name] = method
                                break
                    self._cache[head] = (methods, errors)
                return self._cache[head]
            finally:
                self._thread = None
                self._lock.release()


# =====
def _remove_modules_unsafe(path):
    logger = get_logger()
    logger.debug("Removed modules with path: %s", path)

    for name in list(sys.modules):
        module_path = getattr(sys.modules[name], "__file__", None)
        if module_path is None:
            # FIXME: We don't support the namespaces yet
            logger.debug("Ignored module/package without __file__ attribute: %s", name)
        elif module_path.startswith(os.path.normpath(path) + os.path.sep):
            logger.debug("Removed old module: %s", name)
            del sys.modules[name]


def _get_exposed_unsafe(path):
    assert os.access(path, os.F_OK), "Can't find module path: {}".format(path)

    logger = get_logger()
    logger.debug("Loading rules from path: %s", path)

    sys.path.insert(0, path)
    try:
        modules = {}
        errors = {}
        for name in _get_all_modules(path):
            try:
                modules[name] = importlib.import_module(name)
            except Exception:
                errors[name] = traceback.format_exc()
                logger.exception("Can't import module '%s' from path '%s'", name, path)
        logger.debug("Found %d modules in path '%s'", len(modules), path)

        methods = {}
        for (module_name, module) in modules.items():
            for obj_name in dir(module):
                if obj_name.startswith("__"):
                    continue
                obj = getattr(module, obj_name)
                if callable(obj) and getattr(obj, _ATTR_EXPOSED, False):
                    methods["{}.{}".format(module_name, obj_name)] = obj
        logger.debug("Loaded %d exposed methods from path '%s'", len(methods), path)
        return (methods, errors)
    finally:
        sys.path.remove(path)


def _get_all_modules(base_path):
    base_path = os.path.abspath(base_path)
    make_rel = (lambda root, item: os.path.relpath(os.path.join(root, item), base_path))
    objects = []
    for (root, dirs, files) in os.walk(base_path, followlinks=True):
        if not _is_package_root(base_path, root):
            continue
        for (checker, transformer, items) in (
            (_is_package, (lambda path: path), dirs),
            (_is_module, (lambda path: path[:-3]), files),
        ):
            for item in items:
                if _is_object_name(transformer(item)) and checker(os.path.join(root, item)):
                    obj_path = transformer(make_rel(root, item)).replace(os.path.sep, ".")
                    objects.append(obj_path)
    return sorted(objects)


def _is_package_root(base_path, package_path):
    """
        Проверяет, что переданный package_path на самом деле является пакетом,
        откусывая от его пути base_path:
            package_path = base_path + / + rel_package_path
    """

    rel_package_path = os.path.relpath(package_path, base_path)
    if rel_package_path == ".":
        return True
    path = base_path
    for part in rel_package_path.split(os.path.sep):
        if not _is_object_name(part):
            return False
        path = os.path.join(path, part)
        if not os.path.isfile(os.path.join(path, "__init__.py")):
            return False
    return True


def _is_object_name(name):
    try:
        valid_object_name(name)
        return True
    except ValidatorError:
        return False


def _is_package(path):
    if path.endswith("__pycache__"):
        return False
    return os.path.isfile(os.path.join(path, "__init__.py"))


def _is_module(path):
    return (os.path.basename(path) != "__init__.py" and path.endswith(".py"))



from flask import request

from ulib.validatorlib import ValidatorError
from ulib.validators.extra import valid_hex_string

from .. import tools

from . import (
    ApiError,
    Resource,
)


# =====
class RulesResource(Resource):
    name = "Operations with the rules HEAD"
    methods = ("GET", "POST")
    docstring = """
        GET  -- Returns a current version (head) of the rules in format:

                # =====
                {
                    "status":  "ok",
                    "message": "<...>",
                    "result":  {
                        "head":    "<HEAD>"
                        "errors":  {"<path.to.module>": "<Traceback>", ...}
                        "exposed": {
                            "methods":  ["<path.to.function>", ...],
                            "handlers": ["<path.to.function>", ...],
                        },
                    },
                }
                # =====

                @head    -- Current version of the rules. Null if the version has not yet been set.
                @errors  -- Errors that occurred while loading the specified modules (null if global
                            error occurs).
                @exposed -- Functions loaded from the rules and ready for execution (null if global
                            error occurs).
                @exposed.methods  -- List of functions that can be called directly by name.
                @exposed.handlers -- List of event handlers that are selected based on filters.
                                     They may also be called manually as methods.

                Possible GET errors (with status=="error"):
                    503 -- Non-existant HEAD for rules.

        POST -- Takes a version (head) in the format: {"head": "<HEAD>"} and applies it.

                Return value:
                # =====
                {
                    "status":  "ok",
                    "message": "<...>",
                    "result":  {"head": "<HEAD>"},
                }
                # =====

                Possible POST errors (with status=="error"):
                    400 -- Invalid HEAD (not a hex string).
    """

    def __init__(self, pool, loader):
        self._pool = pool
        self._loader = loader

    def process_request(self):
        with self._pool.get_backend() as backend:
            if request.method == "GET":
                return self._request_get(backend)
            elif request.method == "POST":
                return self._request_post(backend)

    def _request_post(self, backend):
        head = (request.data or {}).get("head")  # json
        try:
            head = valid_hex_string(head)
        except ValidatorError as err:
            raise ApiError(400, str(err), {"head": head})
        backend.rules.set_head(head)
        return ({"head": head}, "The HEAD has been updated")

    def _request_get(self, backend):
        (head, exposed, errors, exc) = tools.get_exposed(backend, self._loader)
        if exc is None:  # No errors
            if exposed is not None:
                exposed_names = {group: list(methods) for (group, methods) in exposed.items()}
            else:
                exposed_names = None  # Not configured HEAD
            return ({"head": head, "exposed": exposed_names, "errors": errors}, "Current HEAD")
        else:
            raise ApiError(503, exc, {"head": head, "exposed": None, "errors": None})



import sys
import os
import multiprocessing
import time

from contextlog import get_logger

from .. import context

from . import init
from . import Application


# =====
_stop = None


def run(args=None, config=None):
    if config is None:
        config = init(__name__, "Powny Worker", args)
    app = _Worker(config)
    global _stop
    _stop = app.stop
    return abs(app.run())


# =====
class _Worker(Application):
    """
        This application performs the jobs. Each job runs in a separate process and
        has its own connection to the backend. It automatically saves the state of
        the job. Worker only ensures that the process must be stopped upon request.
    """

    def __init__(self, config):
        Application.__init__(self, "worker", config)
        self._manager = _JobsManager(self._config.core.rules_dir)
        self._not_started = 0

    def process(self):
        logger = get_logger()
        sleep_mode = False
        with self.get_backend_object().connected() as backend:
            while not self._stop_event.is_set():
                gen_jobs = backend.jobs_process.get_ready_jobs()
                while not self._stop_event.is_set():
                    self._manager.manage(backend)
                    self._write_worker_state(backend)

                    if self._manager.get_current() >= self._app_config.max_jobs:
                        logger.debug("Have reached the maximum concurrent jobs (%d), sleeping %f seconds...",
                                     self._app_config.max_jobs, self._app_config.max_jobs_sleep)
                        time.sleep(self._app_config.max_jobs_sleep)

                    else:
                        try:
                            job = next(gen_jobs)
                        except StopIteration:
                            if not sleep_mode:
                                logger.debug("No jobs in queue, entering to sleep mode with interval %f seconds...",
                                             self._app_config.empty_sleep)
                            sleep_mode = True
                            time.sleep(self._app_config.empty_sleep)
                            break
                        else:
                            sleep_mode = False
                            if not self._manager.run_job(job, self.get_backend_object()):
                                backend.jobs_process.release_job(job.job_id)
                                self._not_started += 1

    def _write_worker_state(self, backend):
        self.set_app_state(backend, {
            "active":      self._manager.get_current(),
            "processed":   self._manager.get_finished(),
            "not_started": self._not_started,
        })


class _JobsManager:
    def __init__(self, rules_dir):
        self._rules_dir = rules_dir
        self._procs = {}
        self._finished = 0

    def get_finished(self):
        return self._finished

    def get_current(self):
        return len(self._procs)

    def run_job(self, job, backend):
        logger = get_logger(job_id=job.job_id, method=job.method_name)
        logger.info("Starting the job process")
        associated = multiprocessing.Event()
        proc = multiprocessing.Process(target=_exec_job, args=(job, self._rules_dir, backend, associated))
        self._procs[job.job_id] = (job.method_name, proc)
        proc.start()
        if not associated.wait(1):
            logger.error("Cannot associate job after one second")
            self._kill(proc)
            return False
        return True

    def manage(self, backend):
        for (job_id, (method_name, proc)) in self._procs.copy().items():
            logger = get_logger(job_id=job_id, method=method_name)
            if not proc.is_alive():
                logger.info("Finished job process %d with retcode %d", proc.pid, proc.exitcode)
                self._finish(job_id)
            elif backend.jobs_process.is_deleted_job(job_id):
                self._kill(proc)
                self._finish(job_id)

    def _finish(self, job_id):
        self._procs.pop(job_id)
        self._finished += 1

    def _kill(self, proc):
        logger = get_logger()
        logger.info("Killing job process %s...", proc)
        try:
            proc.terminate()
            proc.join()
        except Exception:
            logger.exception("Can't kill process %s; ignored", proc)
            return
        logger.info("Killed job process %d with retcode %d", proc.pid, proc.exitcode)


def _exec_job(job, rules_dir, backend, associated):
    logger = get_logger(job_id=job.job_id, method=job.method_name)
    rules_path = os.path.join(rules_dir, job.head)
    with backend.connected():
        logger.debug("Associating job with PID %d", os.getpid())
        backend.jobs_process.associate_job(job.job_id)
        associated.set()

        sys.path.insert(0, rules_path)
        thread = context.JobThread(
            backend=backend,
            job_id=job.job_id,
            state=job.state,
            extra={"number": job.number, "head": job.head, "version": job.head},  # FIXME: version == Compat for YaNS
        )
        thread.start()
        thread.join()



# =====
def make_config(raw, scheme, keys=()):
    if not isinstance(raw, dict):
        raise ValueError("The node '{}' must be a dictionary".format(".".join(keys) or "/"))

    config = Section()
    for (key, option) in scheme.items():
        full_key = keys + (key,)
        full_name = ".".join(full_key)

        if isinstance(option, Option):
            value = raw.get(key, option.default)
            try:
                if value is not None:
                    value = option.type(value)
            except:
                raise ValueError("Invalid value '{value}' for key '{key}'".format(key=full_name, value=value))
            config[key] = value
            config._set_meta(  # pylint: disable=protected-access
                name=key,
                secret=isinstance(option, SecretOption),
                default=option.default,
                help=option.help,
            )
        elif isinstance(option, dict):
            config[key] = make_config(raw.get(key, {}), option, full_key)
        else:
            raise RuntimeError("Incorrect scheme definition for key '{}':"
                               " the value is {}, not dict or [Secret]Option()".format(full_name, type(option)))
    return config


class Section(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self._meta = {}

    def _set_meta(self, name, secret, default, help):  # pylint: disable=redefined-builtin
        self._meta[name] = {
            "secret":  secret,
            "default": default,
            "help":    help,
        }

    def _is_secret(self, name):
        return self._meta[name]["secret"]

    def _get_default(self, name):
        return self._meta[name]["default"]

    def _get_help(self, name):
        return self._meta[name]["help"]

    def __getattribute__(self, name):
        if name in self:
            return self[name]
        else:  # For pickling
            return dict.__getattribute__(self, name)


_type = type


class Option:
    def __init__(self, default, help, type=None):  # pylint: disable=redefined-builtin
        self.default = default
        self.help = help
        self.type = type or (_type(default) if default is not None else str)

    def __repr__(self):
        return "<Option(default={self.default}, type={self.type}, help={self.help})>".format(self=self)


class SecretOption(Option):
    pass



from flask import request

from ulib.validatorlib import ValidatorError
from ulib.validators.extra import valid_uuid

from ..backends import DeleteTimeoutError
from .. import tools

from . import get_url_for
from . import (
    Resource,
    ApiError,
)


# =====
class JobsResource(Resource):
    name = "Create and view jobs"
    methods = ("GET", "POST")
    docstring = """
        GET  -- Returns a dict of all jobs in the system:
                # =====
                {
                    {
                        "status":   "ok",
                        "message":  "<...>",
                        "<job_id>": {
                            "url": "<http://api/url/to/control/the/job>"},
                        },
                        ...
                }
                # =====

        POST -- Run method or handler. If the argument "method" is specified, will
                be running the specified method (requires the full path from the rules).
                If this argument is not specified, it will be found and run the
                appropriate handlers. The arguments passed to method/handler via request
                body (in dict format).

                Return value:
                # =====
                {
                    "status":  "ok",
                    "message": "<...>",
                    "result":  {
                        "<job_id>": {
                            "method": "<path.to.function>",
                            "url": "<http://api/url/to/control/the/job>"},
                        },
                        ...
                    },
                }
                # =====

                For explicit method call will be returnet only one job_id, for handlers
                will be returned several identifiers for the appropriate handlers.

                Possible POST errors (with status=="error"):
                    404 -- Method not found (for method call).
                    503 -- In the queue is more then N jobs.
                    503 -- No HEAD or exposed methods.
    """

    def __init__(self, pool, loader, input_limit):
        self._pool = pool
        self._loader = loader
        self._input_limit = input_limit

    def process_request(self):
        with self._pool.get_backend() as backend:
            if request.method == "GET":
                return self._request_get(backend)
            elif request.method == "POST":
                return self._request_post(backend)

    def _request_get(self, backend):
        result = {
            job_id: {"url": self._get_job_url(job_id)}  # TODO: Add "method" key
            for job_id in backend.jobs_control.get_jobs_list()
        }
        return (result, ("No jobs" if len(result) == 0 else "The list with all jobs"))

    def _request_post(self, backend):
        if backend.jobs_control.get_input_size() >= self._input_limit:
            raise ApiError(503, "In the queue is more then {} jobs".format(self._input_limit))

        (head, exposed) = self._get_exposed(backend)
        method_name = request.args.get("method", None)
        kwargs = dict(request.data or {})

        if method_name is not None:
            result = self._run_method(backend, method_name, kwargs, head, exposed)
            return (result, "Method was launched")
        else:
            result = self._run_handlers(backend, kwargs, head, exposed)
            return (result, ("No matching handler" if len(result) == 0 else "Handlers were launched"))

    def _get_exposed(self, backend):
        (head, exposed, _, _) = tools.get_exposed(backend, self._loader)
        if exposed is None:
            raise ApiError(503, "No HEAD or exposed methods")
        return (head, exposed)

    def _run_method(self, backend, method_name, kwargs, head, exposed):
        state = tools.get_dumped_method(method_name, kwargs, exposed)  # Validation is not required
        if state is None:
            raise ApiError(404, "Method not found")
        job_id = backend.jobs_control.add_job(head, method_name, kwargs, state)
        return {job_id: {"method": method_name, "url": self._get_job_url(job_id)}}

    def _run_handlers(self, backend, kwargs, head, exposed):
        jobs = {}
        for (handler_name, state) in tools.get_dumped_handlers(kwargs, exposed).items():
            job_id = backend.jobs_control.add_job(head, handler_name, kwargs, state)
            jobs[job_id] = {"method": handler_name, "url": self._get_job_url(job_id)}
        return jobs

    def _get_job_url(self, job_id):
        return get_url_for(JobControlResource, job_id=job_id)


class JobControlResource(Resource):
    name = "View and stop job"
    methods = ("GET", "DELETE")
    dynamic = True
    docstring = """
        GET    -- Returns the job state:
                  # =====
                  {
                      "status":  "ok",
                      "message": "<...>",
                      "result":  {
                          "method":   "<path.to.function>",  # Full method path in the rules
                          "head":     "<HEAD>",    # HEAD of the rules for this job
                          "kwargs":   {...},       # Function arguments
                          "number":   <int>,       # The serial number of the job
                          "created":  <str>,       # ISO-8601-like time when the job was created
                          "locked":   <dict|null>, # Job in progress (null if not locked, dict with info otherwise)
                          "deleted":  <str|null>,  # ISO-8601-like time when job was marked to stop and delete
                          "taken":    <str|null>,  # ISO-8601-like time when job was started (taken from queue)
                          "finished": <str|null>,  # ISO-8601-like time when job was finished
                          "stack":    [...],       # Stack snapshot on last checkpoint
                          "retval":   <any|null>,  # Return value if finished and not failed
                          "exc":      <str|null>,  # Text exception if job was failed
                      },
                  }
                  # =====

        DELETE -- Request to immediate stop and remove the job. Waits until the collector does
                  not remove the job.

                  Return value:
                  # =====
                  {
                      "status":  "ok",
                      "message": "<...>",
                      "result": {"deleted": "<job_id>"},
                  }
                  # =====

                  Possible DELETE errors (with status=="error"):
                      503 -- The collector did not have time to remove the job, try again.

        Errors (with status=="error"):
            400 -- Invalid job id.
            404 -- Job not found (or DELETED).
    """

    def __init__(self, pool, delete_timeout):
        self._pool = pool
        self._delete_timeout = delete_timeout

    def process_request(self, job_id):  # pylint: disable=arguments-differ
        with self._pool.get_backend() as backend:
            if request.method == "GET":
                return self._request_get(backend, job_id)
            elif request.method == "DELETE":
                return self._request_delete(backend, job_id)

    def _request_get(self, backend, job_id):
        try:
            job_id = valid_uuid(job_id)
        except ValidatorError as err:
            raise ApiError(400, str(err))
        job_info = backend.jobs_control.get_job_info(job_id)
        if job_info is None:
            raise ApiError(404, "Job not found")
        return (job_info, "Information about the job")

    def _request_delete(self, backend, job_id):
        try:
            deleted = backend.jobs_control.delete_job(job_id, timeout=self._delete_timeout)
        except DeleteTimeoutError as err:
            raise ApiError(503, str(err))
        if not deleted:
            raise ApiError(404, "Job not found")
        else:
            return ({"deleted": job_id}, "The job has been removed")



from .. import tools
from .. import optconf

from . import Resource


# =====
class StateResource(Resource):
    name = "The system statistics"
    docstring = """
        GET -- Returns some information about the system state:
               # =====
               {
                   "status":  "ok",
                   "message": "<...>",
                   "result": {
                       "jobs": {
                           "input":   <number>,
                           "all":     <number>,
                           "counter": <number>,
                       },
                       "apps": {
                           "<app_name>": {
                               "<node_name>" {
                                   "when": "<time>",  # ISO-8601-like when statistics has been writed
                                   "pid":  <int>,
                                   "state": {
                                       "respawns": <int>,  # Number of restarts of the application
                                       ...  # Application-specific fields
                                   },
                               },
                           },
                           ...
                       },
                   },
               }
               # =====

               Application-specific fields:
                   worker:
                       processed -- Number of the processed jobs (exclude active);
                       active    -- Number of the current active jobs.
                   collector:
                       processed -- Number of the processed (removed or pushed-back) jobs.
    """

    def __init__(self, pool):
        self._pool = pool

    def process_request(self):
        with self._pool.get_backend() as backend:
            full_apps_state = backend.system_apps_state.get_full_state()
            full_apps_state.setdefault("worker", {})
            full_apps_state.setdefault("collector", {})
            result = {
                "jobs": {
                    "input":   backend.jobs_control.get_input_size(),
                    "all":     backend.jobs_control.get_jobs_count(),
                    "counter": backend.jobs_control.get_counter_value(),
                },
                "apps": full_apps_state,
            }
            return (result, self.name)


class InfoResource(Resource):
    name = "The system information"
    docstring = """
        GET -- Returns some information about the system in format:
               # =====
               {
                   "status":  "ok",
                   "message": "<...>",
                   "result":  {
                       "version": "<...>",
                       "backend": {
                           "name": "<backend_name>",
                           "info": {...},  # Backen-specific data
                       },
                   },
               }
               # =====
    """

    def __init__(self, pool):
        self._pool = pool

    def process_request(self):
        with self._pool.get_backend() as backend:
            result = {
                "version": tools.get_version(),
                "backend": {
                    "name": self._pool.get_backend_name(),
                    "info": backend.get_info(),
                },
            }
            return (result, self.name)


class ConfigResource(Resource):
    name = "The system configuration"
    docstring = """
        GET -- Returns a dictionary with the system configuration except
               some secret options (passwords, tokens, etc.):
               # =====
               {
                   "status":  "ok",
                   "message": "<...>",
                   "result":  {...},  # Dictionaries
               }
               # =====
    """

    def __init__(self, config):
        self._public = self._make_public(config)

    def process_request(self):
        return (self._public, self.name)

    def _make_public(self, config):
        public = {}
        for (key, value) in config.items():
            if isinstance(value, optconf.Section):
                public[key] = self._make_public(value)
            elif not config._is_secret(key):  # pylint: disable=protected-access
                public[key] = value
        return public



import importlib
import contextlib
import collections
import uuid
from queue import Queue

from contextlog import get_logger


# =====
class DeleteTimeoutError(Exception):
    pass


ReadyJob = collections.namedtuple("ReadyJob", (
    "job_id",
    "number",
    "head",
    "method_name",
    "state",
))


def make_job_id():
    return str(uuid.uuid4())


# =====
class CasNoValueError(Exception):
    pass


class CasVersionError(Exception):
    pass


class CasNoValue:  # Empty value for CAS-storage
    def __new__(cls):
        raise RuntimeError("Use a class rather than an object of class")


CasData = collections.namedtuple("CasData", (
    "value",
    "version",
    "stored",
))


# =====
def get_backend_class(name):
    module = importlib.import_module("powny.backends." + name)
    return getattr(module, "Backend")


class Pool:
    """
        This pool contains several ready-to-use backend objects.
        Some code that wants to use the backend, asks it from the pool.
        After using backend, it's returned to the pool. If the exception occurred,
        backend object will be removed with closing the internal connection and
        replaced to the new object.
    """

    def __init__(self, size, backend_name, backend_opts):
        self._backend_name = backend_name
        self._backend_opts = backend_opts
        self._free_backends = Queue(size)
        for _ in range(size):
            self._free_backends.put(self._create_backend())

    def get_backend_name(self):
        return self._backend_name

    @contextlib.contextmanager
    def get_backend(self):
        backend = self._free_backends.get()
        if not backend.is_connected():
            try:
                backend.open()
            except Exception:
                get_logger().error("Can't open backend %s", backend)
                self._free_backends.put(backend)
                raise
        try:
            yield backend
        finally:
            self._free_backends.put(backend)

    def __len__(self):
        return self._free_backends.qsize()  # Number of free backends

    def _create_backend(self):
        backend_class = get_backend_class(self._backend_name)
        backend = backend_class(**self._backend_opts)
        get_logger().debug("Created backend: %s(%s) = %s", backend_class, self._backend_opts, backend)
        return backend



import contextlib


from . import zoo
from . import ifaces


class Backend:
    """
        ZooKeeper backend provides some interfaces for working with jobs and other operations.
        Each instance of this class contains an own connection to ZooKeeper and can be used for
        any operation.
    """

    def __init__(self, **zoo_kwargs):
        self._client = zoo.Client(**zoo_kwargs)
        self.jobs_control = ifaces.JobsControl(self._client)  # Interface for API
        self.jobs_process = ifaces.JobsProcess(self._client)  # Interface for Worker
        self.jobs_gc = ifaces.JobsGc(self._client)  # Interface for Collector
        self.rules = ifaces.Rules(self._client)  # API and internal interface to control the rules
        self.system_apps_state = ifaces.AppsState(self._client)  # API and internal interface to the system statistics
        self.cas_storage = ifaces.CasStorage(self._client)  # Basic CAS storage for user scripts

    @classmethod
    def get_options(cls):
        return zoo.Client.get_options()

    # ===

    @contextlib.contextmanager
    def connected(self):
        self.open()
        try:
            yield self
        finally:
            self.close()

    def open(self):
        self._client.open()
        ifaces.init(self._client)

    def close(self):
        self._client.close()

    def is_connected(self):
        return self._client.is_connected()

    # ===

    def get_info(self):
        return self._client.get_server_info()



from contextlog import get_logger


# =====
_ATTR_ON_EVENT = "_powny_on_event"
_ATTR_MATCHERS = "_powny_matchers"


# =====
def on_event(method):
    setattr(method, _ATTR_ON_EVENT, True)
    return method


def is_event_handler(method):
    return getattr(method, _ATTR_ON_EVENT, False)


def match_event(*matchers):
    assert len(matchers) > 0, "Required minimum one matcher"

    def decorator(method):
        method_matchers = getattr(method, _ATTR_MATCHERS, [])
        for matcher in matchers:
            method_matchers.append(matcher)
        setattr(method, _ATTR_MATCHERS, method_matchers)
        return method

    return decorator


def check_match(method, event):
    for matcher in getattr(method, _ATTR_MATCHERS, []):
        logger = get_logger(method="{}.{}".format(method.__module__, method.__name__))
        try:
            if not matcher(event):
                logger.debug("Event is not matched by %s; data: %s", matcher, event)
                return False
        except Exception:
            logger.exception("Matching error with matcher %s; data: %s", matcher, event)
            return False
    return True



import setuptools


# =====
if __name__ == "__main__":
    setuptools.setup(
        name="powny",
        version="1.3.0",
        url="https://github.com/yandex-sysmon/powny",
        license="GPLv3",
        author="Devaev Maxim",
        author_email="mdevaev@gmail.com",
        description="Distributed events processor, based on stackless technology of PyPy3",
        platforms="any",

        packages=[
            "powny",
            "powny.core",
            "powny.core.api",
            "powny.core.apps",
            "powny.core.optconf",  # TODO: Make a separate package
            "powny.core.optconf.loaders",
            "powny.backends",
            "powny.backends.zookeeper",
            "powny.testing",
        ],

        package_data={
            "powny.core.api": ["templates/*.html"],
            "powny.core.apps": ["configs/*.yaml"],
        },

        entry_points={
            "console_scripts": [
                "powny-api = powny.core.apps.api:run",
                "powny-worker = powny.core.apps.worker:run",
                "powny-collector = powny.core.apps.collector:run",
            ]
        },

        classifiers=[  # http://pypi.python.org/pypi?:action=list_classifiers
            "Development Status :: 4 - Beta",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: Implementation :: PyPy",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: System :: Distributed Computing",
            "Topic :: System :: Networking :: Monitoring",
        ],

        install_requires=[
            "kazoo.yandex >=2.0",  # https://github.com/python-zk/kazoo/pull/252
            "Flask-API.yandex",
            "python-dateutil",
            "gunicorn",
            "pyyaml",
            "ulib",
            "decorator",
            "contextlog",
            "colorlog",
            "pkginfo",

            # Backdoor
            "manhole",

            # Optconf
            "tabloid",
            "colorama",
            "pygments",
        ],
    )
